from __future__ import annotations

import time
from abc import ABC
from dataclasses import InitVar, dataclass, field
from typing import Dict, List

import seacharts.environment.scope as env

from .base import Layer
from .layers import Land, Seabed, Shore


@dataclass
class _Hypsometry(ABC):
    scope: InitVar[env.Scope]

    def __post_init__(self, scope: env.Scope):
        raise NotImplementedError

    @property
    def layers(self) -> List[Layer]:
        raise NotImplementedError

    def add_buffer(self, layer, distance):
        raise NotImplementedError

    @property
    def loaded_layers(self) -> List[Layer]:
        return [layer for layer in self.layers if not layer.geometry.is_empty]

    def load(self, scope: env.Scope):
        layers = [x for x in self.layers if x.label in scope.layers]
        if scope.new_data:
            if scope.parser.verbose:
                print(f"Processing {scope.extent.area // 10 ** 6} km^2 of " f"{self.__class__.__name__} features from: " + ", ".join(scope.files))
            for layer in layers:
                start_time = time.time()
                records = layer.load_fgdb(scope.parser)
                info = f"{len(records)} {layer.name} geometries"

                if not records:
                    if scope.parser.verbose:
                        print(f"\rFound {info}.\n")
                    return

                if scope.raw_data:
                    if scope.parser.verbose:
                        print(f"\rExtracting raw data from {info}...", end="")
                    layer.extract_raw(records)
                else:
                    if scope.parser.verbose:
                        print(f"\rMerging {info}...", end="")
                    layer.unify(records)

                    if scope.parser.verbose:
                        print(f"\rSimplifying {info}...", end="")
                    layer.simplify(scope.tolerance)

                    if scope.parser.verbose:
                        print(f"\rBuffering {info}...", end="")
                    self.add_buffer(layer, scope.buffer)

                    if scope.parser.verbose:
                        print(f"\rClipping {info}...", end="")
                    layer.clip(scope.extent.bbox)

                layer.save(scope.parser)
                if scope.parser.verbose:
                    end_time = round(time.time() - start_time, 1)
                    print(f"\rSaved {info} to shapefile in {end_time} seconds.")

            if scope.parser.verbose:
                print()
        else:
            print("ENC spatial.hypsometry.load Warning: Using previously loaded data. Use new-data=True to reload.")
            for layer in layers:
                layer.load_shapefile(scope.parser)


@dataclass
class Hydrography(_Hypsometry):
    bathymetry: Dict[int, Layer] = field(init=False)

    @property
    def layers(self):
        return [*self.bathymetry.values()]

    def __post_init__(self, scope: env.Scope):
        self.bathymetry = {d: Seabed(d) for d in scope.depths}
        self.load(scope)

    def add_buffer(self, layer, distance):
        layer.erode(distance)


@dataclass
class Topography(_Hypsometry):
    land: Land = field(init=False)
    shore: Shore = field(init=False)

    @property
    def layers(self):
        return [self.land, self.shore]

    def __post_init__(self, scope: env.Scope):
        self.land = Land()
        self.shore = Shore()
        self.load(scope)

    def add_buffer(self, layer, distance):
        layer.dilate(distance)



from __future__ import annotations

import seacharts.spatial as spl

from .extent import Extent
from .scope import Scope


class Environment:
    supported_layers = ", ".join(spl.supported_layers)

    def __init__(self, settings: dict):
        self.supported_crs = "EUREF89 UTM zone " + str(settings["enc"]["utm_zone"])
        extent = Extent(settings)
        self.scope = Scope(settings, extent)
        self.hydrography = spl.Hydrography(self.scope)
        self.topography = spl.Topography(self.scope)
        self.safe_area = None
        self.ownship = None
        self.depth = None

    def create_ownship(self, x, y, heading, hull_scale, lon_scale, lat_scale) -> None:
        self.ownship = spl.Ship(
            x,
            y,
            heading,
            scale=hull_scale,
            lon_scale=lon_scale,
            lat_scale=lat_scale,
        )

    def filter_hazardous_areas(self, depth, buffer=0) -> None:
        if not isinstance(depth, int) or depth not in self.scope.depths:
            raise ValueError("Danger area depth must be an integer from chosen depths: " f"{self.scope.depths}")
        self.depth = depth
        if buffer < 0:
            raise ValueError("Buffer should be a positive integer.")
        self.safe_area = self.hydrography.bathymetry[depth]
        if buffer:
            self.safe_area.erode(buffer)



if __name__ == "__main__":
    import shapely.geometry as geo
    from seacharts.enc import ENC

    size = 9000, 5062
    center = 44300, 6956450
    enc = ENC(new_data=True, size=size, center=center)
    enc.start_display()

    # (id, easting, northing, heading, color)
    ships = [
        (1, 46100, 6957000, 132, "orange"),
        (2, 45000, 6956000, 57, "yellow"),
        (3, 44100, 6957500, 178, "red"),
        (4, 42000, 6955200, 86, "green"),
        (5, 44000, 6955500, 68, "pink"),
    ]

    enc.add_vessels(*ships)

    x, y = center
    width, height = 1900, 1900
    box = geo.Polygon(
        (
            (x - width, y - height),
            (x + width, y - height),
            (x + width, y + height),
            (x - width, y + height),
        )
    )
    areas = list(box.difference(enc.seabed[10].geometry).geoms)
    enc.draw_circle(center, 1000, "yellow", thickness=2, edge_style="--", alpha=0.5)
    enc.draw_rectangle(center, (600, 1200), "blue", rotation=20, alpha=0.5)
    enc.draw_circle(
        center, 700, "green", edge_style=(0, (5, 8)), thickness=3, fill=False
    )
    enc.draw_line(
        [(center[0], center[1] + 800), center, (center[0] - 300, center[1] - 400)],
        "white",
    )
    enc.draw_line(
        [
            (center[0] - 300, center[1] + 400),
            center,
            (center[0] + 200, center[1] - 600),
        ],
        "magenta",
        width=0.0,
        thickness=5.0,
        edge_style=(0, (1, 4)),
    )
    enc.draw_arrow(
        center,
        (center[0] + 700, center[1] + 600),
        "orange",
        head_size=300,
        width=50,
        thickness=5,
    )
    enc.draw_polygon(enc.seabed[100].geometry.geoms[-3], "cyan", alpha=0.5)
    enc.draw_polygon(enc.shore.geometry.geoms[56], "highlight", alpha=0.5)
    for area in areas[3:8] + [areas[14], areas[17]] + areas[18:21]:
        enc.draw_polygon(area, "red", alpha=0.5)
    enc.draw_rectangle(
        center,
        (width, height),
        "pink",
        fill=False,
        edge_style=(0, (10, 10)),
        thickness=1.5,
    )

    enc.show_display()  # Show display indefinitely



from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple


@dataclass
class Extent:
    def __init__(self, settings: dict):
        self.utm_zone: int = settings["enc"]["utm_zone"]
        self.size: Tuple[int, int] = settings["enc"]["size"]
        self.size = self.size[0], self.size[1]

        self.origin: Tuple[int, int] = settings["enc"]["origin"]
        self.origin = self.origin[0], self.origin[1]

        if "center" in settings["enc"]:
            self.center: Tuple[int, int] = settings["enc"]["center"]
            self.center = self.center[0], self.center[1]
            self._origin_from_center()
        else:
            self._center_from_origin()

        self.bbox: Tuple[int, int, int, int] = self._bounding_box_from_origin_size()
        self.area: int = self.size[0] * self.size[1]

    def _origin_from_center(self) -> None:
        self.origin = (self.center[0] - self.size[0] / 2.0, self.center[1] - self.size[1] / 2.0)

    def _center_from_origin(self) -> None:
        self.center = (self.origin[0] + self.size[0] / 2.0, self.origin[1] + self.size[1] / 2.0)

    def _bounding_box_from_origin_size(self) -> Tuple[int, int, int, int]:
        x_min, y_min = self.origin
        x_max, y_max = x_min + self.size[0], y_min + self.size[1]
        return x_min, y_min, x_max, y_max



from __future__ import annotations

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import seacharts.display as dis

matplotlib.rcParams["pdf.fonttype"] = 42
matplotlib.rcParams["ps.fonttype"] = 42


class EventsManager:
    _zoom_scale = 0.9
    _heading_step = 5
    _cartesian_step = 50
    _directions = {"up": 1, "down": -1, "left": -1, "right": 1}
    _resizing = {",": 0.9, ".": 1.1, "[": 0.9, "]": 1.1}

    def __init__(self, display: dis.Display):
        self._display = display
        self._canvas = display.figure.canvas
        self._view_limits = dict(x=None, y=None)
        self._direction_keys = {d: False for d in self._directions}
        self._control_pressed = False
        self._shift_pressed = False
        self._mouse_press = None
        self._connect_canvas_events()
        self._remove_default_keybindings()

    def _connect_canvas_events(self):
        self._canvas.mpl_connect("scroll_event", self._handle_zoom)
        self._canvas.mpl_connect("key_press_event", self._key_press)
        self._canvas.mpl_connect("key_release_event", self._key_release)
        self._canvas.mpl_connect("button_press_event", self._click_press)
        self._canvas.mpl_connect("button_release_event", self._click_release)
        self._canvas.mpl_connect("motion_notify_event", self._mouse_motion)

    def _handle_zoom(self, event):
        if event.button == "down":
            scale_factor = 1 / self._zoom_scale
        elif event.button == "up":
            scale_factor = self._zoom_scale
        else:
            raise NotImplementedError
        x_limit = self._display.axes.get_xlim()
        y_limit = self._display.axes.get_ylim()
        new_width = (x_limit[1] - x_limit[0]) * scale_factor
        new_height = (y_limit[1] - y_limit[0]) * scale_factor
        x_data, y_data = event.xdata, event.ydata
        dx = (x_limit[1] - x_data) / (x_limit[1] - x_limit[0])
        dy = (y_limit[1] - y_data) / (y_limit[1] - y_limit[0])
        self._display.axes.set_xlim([x_data - new_width * (1 - dx), x_data + new_width * dx])
        self._display.axes.set_ylim([y_data - new_height * (1 - dy), y_data + new_height * dy])
        self._display.draw_plot()

    def _key_press(self, event):
        if event.key == "escape":
            self._display.terminate()
        elif event.key == "d":
            self._display.toggle_dark_mode()
        elif event.key == "t":
            self._display.features.show_top_hidden_layer()
        elif event.key == "g":
            self._display.features.show_bottom_hidden_layer()
        elif event.key == "h":
            self._display.features.hide_top_visible_layer()
        elif event.key == "b":
            self._display.features.hide_bottom_visible_layer()
        elif event.key == "u":
            self._display.features.update_vessels()
            self._display.update_plot()
        elif event.key == "v":
            self._display.features.toggle_vessels_visibility()
        elif event.key == "l":
            self._display.features.toggle_topography_visibility()
        elif event.key in ["n", "m"]:
            if self._display.environment.ownship and self._display.features.show_ownship and self._display.environment.depth is not None:
                self._toggle_depth_filter_values(event.key)
        elif event.key == "o":
            if self._display.environment.ownship:
                self._display.features.toggle_ownship_visibility()
            else:
                self._add_ownship_to_plot_center()
        elif event.key == "z":
            self._display.features.toggle_hazards_visibility()
        elif event.key == "a":
            self._display.features.toggle_arrows_visibility()
        elif event.key == "c":
            self._display.toggle_colorbar()
        elif event.key == "f":
            self._display.toggle_fullscreen()
        elif event.key == "ctrl+s":
            self._display.save_figure("svg", extension="svg")
        elif event.key == "s":
            self._display.save_figure("low_res", 2.0)
        elif event.key == "S":
            self._display.save_figure("high_res", 10.0)
        elif event.key in self._directions:
            if self._display.environment.ownship and self._display.features.show_ownship:
                self._direction_keys[event.key] = True
                self._move_ownship()
        elif event.key in self._resizing:
            if self._display.environment.ownship and self._display.features.show_ownship:
                self._resize_hazards_horizon(event.key)
        elif event.key == "shift":
            self._shift_pressed = True
        elif event.key == "control":
            self._control_pressed = True
        elif event.key[:4] == "alt+":
            key = event.key[4:]
            if key in self._directions:
                self._move_figure_position(key)

    def _key_release(self, event):
        if event.key in self._directions:
            self._direction_keys[event.key] = False
        elif event.key == "shift":
            self._shift_pressed = False
        elif event.key == "control":
            self._control_pressed = False

    def _move_figure_position(self, key):
        matrix = self._display.window_anchors
        j, i = self._display.anchor_index
        if key == "left" or key == "right":
            i = (i + self._directions[key]) % len(matrix[0])
        elif key == "up" or key == "down":
            j = (j - self._directions[key]) % len(matrix)
        self._display.anchor_index = j, i
        self._display.set_figure_position()

    def _add_ownship_to_plot_center(self):
        center = self._display.environment.scope.extent.center
        x, y = center[0] - 1414, center[1] - 1414
        self._display.environment.create_ownship(x, y, 45, 1, 10, 10)
        self._display.environment.filter_hazardous_areas(10)
        self._display.features.toggle_ownship_visibility()

    def _move_ownship(self):
        x, y, psi, *params = self._display.environment.ownship.parameters
        if self._direction_keys["up"]:
            x = x + np.sin(np.deg2rad(psi)) * self._cartesian_step
            y = y + np.cos(np.deg2rad(psi)) * self._cartesian_step
        if self._direction_keys["down"]:
            x = x - np.sin(np.deg2rad(psi)) * self._cartesian_step
            y = y - np.cos(np.deg2rad(psi)) * self._cartesian_step
        if self._direction_keys["left"]:
            psi = psi - self._heading_step
        if self._direction_keys["right"]:
            psi = psi + self._heading_step
        self._display.environment.create_ownship(x, y, psi, *params)
        self._display.features.update_ownship()
        self._display.features.update_hazards()
        self._display.update_plot()

    def _resize_hazards_horizon(self, key):
        *other, lon, lat = self._display.environment.ownship.parameters
        if key == "," or key == ".":
            lon *= self._resizing[key]
        else:
            lat *= self._resizing[key]
        self._display.environment.create_ownship(*other, lon, lat)
        self._display.features.update_ownship()
        self._display.features.update_hazards()
        self._display.update_plot()

    def _toggle_depth_filter_values(self, key):
        d = self._display.environment.depth
        d_index = self._display.environment.scope.depths.index(d)
        delta = 1 if key == "m" else -1
        next_d = d_index + delta
        if 0 <= next_d < len(self._display.environment.scope.depths):
            d = self._display.environment.scope.depths[next_d]
            self._display.environment.depth = d
            self._display.environment.filter_hazardous_areas(d)
            self._display.features.update_hazards()
            self._display.update_plot()

    def _click_press(self, event):
        if event.inaxes != self._display.axes:
            return
        if self._shift_pressed or self._control_pressed:
            if event.button in [plt.MouseButton.LEFT, plt.MouseButton.RIGHT]:
                pick = event.xdata, event.ydata
                coords = pick if event.button == plt.MouseButton.LEFT else None
                path = 1 if self._shift_pressed else 2
                self._mouse_press = dict(x=pick[0], y=pick[1])
                self._display.features.update_waypoints(path, pick, coords)
                self._display.update_plot()
        else:
            if event.button == plt.MouseButton.LEFT:
                self._view_limits["x"] = self._display.axes.get_xlim()
                self._view_limits["y"] = self._display.axes.get_ylim()
                self._mouse_press = dict(x=event.xdata, y=event.ydata)

    def _click_release(self, _):
        self._mouse_press = None
        self._display.draw_plot()

    def _mouse_motion(self, event):
        if self._mouse_press is None:
            return
        if event.inaxes != self._display.axes:
            return
        if self._shift_pressed or self._control_pressed:
            new_pos = event.xdata, event.ydata
            old_pos = self._mouse_press["x"], self._mouse_press["y"]
            path = 1 if self._shift_pressed else 2
            self._display.features.update_waypoints(path, old_pos, new_pos)
            self._mouse_press = dict(x=new_pos[0], y=new_pos[1])
            self._display.update_plot()
        else:
            self._view_limits["x"] -= event.xdata - self._mouse_press["x"]
            self._view_limits["y"] -= event.ydata - self._mouse_press["y"]
            self._display.axes.set_xlim(self._view_limits["x"])
            self._display.axes.set_ylim(self._view_limits["y"])
            self._display.draw_plot()

    @staticmethod
    def _remove_default_keybindings():
        dic = plt.rcParams
        # dic['keymap.pan'].remove('p')
        # dic["keymap.zoom"].remove("o")
        # dic["keymap.grid"].remove("g")
        # dic["keymap.save"].remove("s")
        # dic["keymap.save"].remove("ctrl+s")
        # dic["keymap.back"].remove("left")
        # dic["keymap.xscale"].remove("k")
        # dic["keymap.yscale"].remove("l")
        # dic["keymap.forward"].remove("v")
        # dic["keymap.forward"].remove("right")
        # # dic['keymap.all_axes'].remove('a')
        # dic["keymap.fullscreen"].remove("f")



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field
from typing import Any, List

from shapely import geometry as geo
from shapely import ops


@dataclass
class Drawable:
    scale: float = field(init=False, repr=False)
    color: str = field(init=False, repr=False)
    z_order: int = field(init=False, repr=False)
    artist: Any = field(init=False, repr=False)


@dataclass
class Coordinates:
    x: float
    y: float


@dataclass
class Vector(Coordinates):
    pass


@dataclass
class Radial:
    radius: float


@dataclass
class Oriented:
    heading: float
    in_degrees: bool = True


@dataclass
class ZeroDepth:
    depth = 0


@dataclass
class SingleDepth:
    depth: int


@dataclass
class MultiDepth:
    @property
    def depth(self):
        raise AttributeError("Multi-depth shapes have no single depth.")


@dataclass
class Shape(Drawable, ABC):
    geometry: geo.base.BaseGeometry = None

    def simplify(self, tolerance, preserve_topology=True):
        self.geometry = self.geometry.simplify(tolerance, preserve_topology)

    def clip(self, bbox):
        bounding_box = geo.box(*bbox)
        self.geometry = bounding_box.intersection(self.geometry)

    def dilate(self, distance):
        self.geometry = self.geometry.buffer(distance, cap_style=2, join_style=3)

    def erode(self, distance):
        self.dilate(-distance)

    def merge(self, other: Shape):
        self.geometry = self.geometry.union(other.geometry)

    def closest_points(self, geometry):
        return ops.nearest_points(self.geometry, geometry)[1]

    @property
    def mapping(self) -> dict:
        return geo.mapping(self.geometry)

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def invalid(self):
        return not self.geometry.is_valid

    @staticmethod
    def is_multi(geometry):
        return isinstance(geometry, geo.MultiPolygon) or isinstance(geometry, geo.GeometryCollection)

    @staticmethod
    def _record_to_geometry(record):
        return geo.shape(record["geometry"])

    @staticmethod
    def as_multi(geometry):
        if isinstance(geometry, geo.Point):
            return geo.MultiPoint([geometry])
        elif isinstance(geometry, geo.Polygon):
            return geo.MultiPolygon([geometry])
        elif isinstance(geometry, geo.LineString):
            return geo.MultiLineString([geometry])
        else:
            raise NotImplementedError(type(geometry))

    @staticmethod
    def collect(geometries):
        if any(not g.is_valid for g in geometries):
            geometries = [g.buffer(0) if not g.is_valid else g for g in geometries]
        geometry = ops.unary_union(geometries)
        if not geometry.is_valid:
            geometry = geometry.buffer(0)
        return geometry

    @staticmethod
    def line_between(point1, point2):
        return geo.LineString([point1, point2])

    @staticmethod
    def arrow_head(points):
        return geo.Polygon(points)


@dataclass
class Layer(Shape, ABC):
    @property
    def _external_labels(self) -> List[str]:
        raise NotImplementedError

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def label(self) -> str:
        return self.name.lower()

    def save(self, parser):
        parser.write(self)

    def load_shapefile(self, parser):
        records = list(parser.read_shapefile(self.label))
        if len(records) > 0:
            self.geometry = self._record_to_geometry(records[0])
            if isinstance(self.geometry, geo.Polygon):
                self.geometry = self.as_multi(self.geometry)

    def load_fgdb(self, parser):
        depth = self.depth if hasattr(self, "depth") else 0
        return list(parser.read_fgdb(self.label, self._external_labels, depth))

    def unify(self, records):
        geometries = [self._record_to_geometry(r) for r in records]
        self.geometry = self.collect(geometries)

    def extract_raw(self, records):
        geometries = [self._record_to_geometry(r) for r in records]
        self.geometry = geo.MultiPolygon(
            [list(g)[0] if isinstance(g, geo.MultiPolygon) else g.buffer(1) for g in geometries]
        )


@dataclass
class Locations(Layer, ABC):
    geometry: geo.MultiPoint = field(default_factory=geo.MultiPoint)


@dataclass
class ZeroDepthLocations(Locations, ZeroDepth, ABC):
    pass


@dataclass
class SingleDepthLocations(Locations, SingleDepth, ABC):
    pass


@dataclass
class MultiDepthLocations(Locations, MultiDepth, ABC):
    pass


@dataclass
class Regions(Layer, ABC):
    geometry: geo.MultiPolygon = field(default_factory=geo.MultiPoint)


@dataclass
class ZeroDepthRegions(Regions, ZeroDepth, ABC):
    pass


@dataclass
class SingleDepthRegions(Regions, SingleDepth, ABC):
    @property
    def name(self):
        return self.__class__.__name__ + f"{self.depth}m"


@dataclass
class MultiDepthRegions(Regions, MultiDepth, ABC):
    pass



from __future__ import annotations

import datetime
import time
import tkinter as tk
from multiprocessing import Process
from pathlib import Path
from typing import List, Tuple

import matplotlib
import matplotlib.pyplot as plt
from cartopy.crs import UTM
from matplotlib.gridspec import GridSpec
from matplotlib_scalebar.scalebar import ScaleBar

import seacharts.environment as env
from .colors import colorbar
from .events import EventsManager
from .features import FeaturesManager

matplotlib.rcParams["pdf.fonttype"] = 42
matplotlib.rcParams["ps.fonttype"] = 42


class Display:
    crs = None
    window_anchors = (
        ("top_left", "top", "top_right"),
        ("left", "center", "right"),
        ("bottom_left", "bottom", "bottom_right"),
    )

    def __init__(self, settings: dict, environment: env.Environment):
        self.environment = environment
        self._show_figure = settings["display"]["show_figure"]
        self._utm_zone = settings["enc"]["utm_zone"]
        self._setup_figure_parameters(settings)

    def _setup_figure_parameters(self, settings: dict) -> None:
        if self._show_figure:
            self.crs = UTM(settings["enc"]["utm_zone"])
            self._background = None
            self.anchor_index = self._init_anchor_index(settings)
            self.figure, self.sizes, self.spacing, widths = self._init_figure(settings)
            self.axes, self.grid_spec, self._colorbar = self._init_axes(widths)
            self.events = EventsManager(self)
            self.features = FeaturesManager(self)
            self.axes.add_artist(
                ScaleBar(
                    1,
                    units="m",
                    location="lower left",
                    frameon=False,
                    color="white",
                    box_alpha=0.0,
                    pad=0.5,
                    font_properties={"size": 12},
                )
            )

            self.draw_plot()

            if self._fullscreen_mode:
                self.toggle_fullscreen()
            else:
                self.set_figure_position()

            # if self._colorbar_mode:
            #     self.toggle_colorbar()

            self.toggle_dark_mode(self._dark_mode)

            if self.environment is None:
                self.start_visualization_loop()

    def start(self, settings: dict) -> None:
        """Starts the display, if it is not already started.
        Overrides the show_figure setting if set to false.

        Args:
            settings (dict): The ENC settings dictionary.
        """
        if self.is_active:
            return

        if not self._show_figure:
            self._show_figure = True

        self._setup_figure_parameters(settings)
        plt.show(block=False)

    def _init_anchor_index(self, settings):
        option = settings["display"]["anchor"]
        for j, window_anchor in enumerate(self.window_anchors):
            if option in window_anchor:
                return j, window_anchor.index(option)
        raise ValueError(
            f"Invalid window anchor option '{option}', "
            f"possible candidates are: \n"
            f"{[o for options in self.window_anchors for o in options]}"
        )

    def _init_figure(self, settings):
        self._fullscreen_mode = settings["display"]["fullscreen_mode"]
        self._colorbar_mode = settings["display"]["colorbar_mode"]
        self._dark_mode = settings["display"]["dark_mode"]
        self._dpi = settings["display"]["dpi"]
        self._resolution = settings["display"]["resolution"]

        if self._fullscreen_mode:
            plt.rcParams["toolbar"] = "None"

        width, height = self.environment.scope.extent.size
        window_height, ratio = self._resolution / self._dpi, width / height
        figure_width1, figure_height1 = ratio * window_height, window_height
        axes1_width, axes2_width, width_space = figure_width1, 1.1, 0.3
        axes_widths = axes1_width, axes2_width
        figure_height2 = figure_height1 * 0.998
        figure_width2 = axes1_width + width_space + 2 * axes2_width
        figure_sizes = [
            (figure_width1, figure_height1),
            (figure_width2, figure_height2),
        ]
        sub1 = dict(
            right=(axes1_width + width_space + axes2_width) / figure_width1,
            wspace=2 * width_space / (axes1_width + axes2_width),
        )
        sub2 = dict(
            right=(axes1_width + width_space + axes2_width) / figure_width2,
            wspace=2 * width_space / axes1_width,
        )
        subplot_spacing = sub1, sub2
        figure = plt.figure("SeaCharts", figsize=figure_sizes[0], dpi=self._dpi)
        return figure, figure_sizes, subplot_spacing, axes_widths

    def _init_axes(self, axes_widths):
        gs = GridSpec(
            1,
            2,
            width_ratios=axes_widths,
            **self.spacing[0],
            left=0.0,
            top=1.0,
            bottom=0.0,
            hspace=0.0,
        )
        axes1 = self.figure.add_subplot(gs[0, 0], projection=self.crs)
        x_min, y_min, x_max, y_max = self.environment.scope.extent.bbox
        axes1.set_extent((x_min, x_max, y_min, y_max), crs=self.crs)
        axes2 = self.figure.add_subplot(gs[0, 1])
        cb = colorbar(axes2, self.environment.scope.depths)
        return axes1, gs, cb

    def start_visualization_loop(self):
        if not self._show_figure:
            return

        print()
        self.show(0.1)
        start_time = time.time()
        while True:
            delta = datetime.timedelta(seconds=time.time() - start_time)
            now = str(delta).split(".")[0]
            print(f"\rVisualizing multiprocessing environment | {now}", end="")
            if not self.is_active:
                self.terminate()
                print()
                return
            self.features.update_vessels()
            self.update_plot()
            time.sleep(0.1)

    def refresh_vessels(self, poses: List[Tuple]):
        if not self._show_figure:
            return

        self.features.vessels_to_file(poses)
        self.features.update_vessels()
        self.update_plot()

    def update_plot(self):
        if not self._show_figure:
            return

        self.figure.canvas.restore_region(self._background)
        self.draw_animated_artists()

    def draw_plot(self):
        if not self._show_figure:
            return
        plt.show(block=False)
        try:
            self.figure.canvas.draw()
        except tk.TclError:
            plt.close()
        self._background = self.figure.canvas.copy_from_bbox(self.figure.bbox)
        self.draw_animated_artists()

    def draw_animated_artists(self):
        if not self._show_figure:
            return

        for artist in self.features.animated:
            self.axes.draw_artist(artist)
        try:
            self.figure.canvas.blit()
            self.figure.canvas.flush_events()
        except tk.TclError:
            plt.close()

    def toggle_dark_mode(self, state=None):
        if not self._show_figure:
            return

        state = state if state is not None else not self._dark_mode
        color = "#142c38" if state else "#ffffff"
        self.figure.set_facecolor(color)
        self.figure.axes[0].set_facecolor(color)
        self._colorbar.ax.set_facecolor(color)
        self.features.toggle_topography_visibility(not state)
        self._dark_mode = state
        self.draw_plot()

    def toggle_colorbar(self, state=None):
        if not self._show_figure:
            return

        if state is not None:
            self._colorbar_mode = state
        else:
            self._colorbar_mode = not self._colorbar_mode
        self.grid_spec.update(**self.spacing[int(self._colorbar_mode)])
        if not self._fullscreen_mode:
            self.figure.set_size_inches(self.sizes[int(self._colorbar_mode)])
            self.set_figure_position()
        self.draw_plot()

    def toggle_fullscreen(self, state=None):
        if not self._show_figure:
            return

        if state is not None:
            self._fullscreen_mode = state
        else:
            self._fullscreen_mode = not self._fullscreen_mode
        plt.get_current_fig_manager().full_screen_toggle()
        if not self._fullscreen_mode:
            self.figure.set_size_inches(self.sizes[int(self._colorbar_mode)])
            self.set_figure_position()
        self.draw_plot()

    def set_figure_position(self):
        if not self._show_figure:
            return

        j, i = self.anchor_index
        option = self.window_anchors[j][i]
        if option != "default":
            root = tk.Tk()
            screen_width = int(root.winfo_screenwidth())
            screen_height = int(root.winfo_screenheight())
            root.destroy()
            x_margin, y_margin = -10, -73
            dpi = self._dpi
            size = self.sizes[int(self._colorbar_mode)]
            width, height = [int(size[k] * dpi) for k in range(2)]
            x_shifted = screen_width - width
            y_shifted = screen_height - height
            if option == "center":
                x, y = x_shifted // 2, y_shifted // 2
            elif option == "right":
                x, y = x_shifted, y_shifted // 2
            elif option == "left":
                x, y = 4, y_shifted // 2
            elif option == "top":
                x, y = x_shifted // 2, 2
            elif option == "bottom":
                x, y = x_shifted // 2, y_shifted + y_margin
            elif option == "top_right":
                x, y = x_shifted, 2
            elif option == "top_left":
                x, y = 4, 2
            elif option == "bottom_right":
                x, y = x_shifted, y_shifted + y_margin
            elif option == "bottom_left":
                x, y = 4, y_shifted + y_margin
            else:
                x, y = 4, 2
            manager = plt.get_current_fig_manager()
            manager.window.wm_geometry(f"+{x + x_margin}+{y}")

    def save_figure(
        self,
        name: str | None = None,
        path: Path | None = None,
        scale: float = 1.0,
        extension: str = "png",
    ):
        if not self._show_figure:
            return

        try:
            if name is None:
                name = self.figure.canvas.manager.get_window_title()

            if path is None:
                path_str = f"reports/{name}.{extension}"
            else:
                path_str = str(path / f"{name}.{extension}")

            self.figure.savefig(
                path_str,
                dpi=self.figure.dpi * scale,
                bbox_inches=self.figure.bbox_inches,
                pad_inches=0.0,
            )
        except tk.TclError:
            plt.close()

    @property
    def utm_zone(self):
        return self._utm_zone

    @property
    def is_active(self):
        if not self._show_figure:
            return

        return plt.fignum_exists(self.figure.number)

    def show(self, duration=0.0):
        if not self._show_figure:
            return

        if self.environment.ownship:
            self.features.update_ownship()
            if self.environment.safe_area:
                self.features.update_hazards()
        try:
            plt.pause(duration)
        except tk.TclError:
            plt.close()

    def terminate(self):
        if not self._show_figure:
            return
        plt.close(self.figure)
        self._show_figure = False

    @staticmethod
    def init_multiprocessing():
        Process(target=Display).start()



from __future__ import annotations

import seacharts.display as dis
import seacharts.spatial as spl
import seacharts.utils as utils
import shapely.geometry as sgeo
from cartopy.feature import ShapelyFeature

from .colors import color_picker


class FeaturesManager:
    def __init__(self, display: dis.Display):
        self._display = display
        self.show_ownship = False
        self.show_hazards = False
        self.show_vessels = True
        self.show_arrows = True
        self._paths = [spl.Path("yellow"), spl.Path("pink")]
        self._ownship = None
        self._horizon = None
        self._vessels = {}
        self._hazards = {}
        self._arrows = {}
        self._seabeds = {}
        self._land = None
        self._shore = None
        self._init_layers()

    @property
    def animated(self):
        return [
            a
            for a in [
                self._horizon,
                *self._hazards.values(),
                *self._arrows.values(),
                self._paths[0].artist,
                self._paths[1].artist,
                *[v["artist"] for v in self._vessels.values()],
                self._ownship,
            ]
            if a
        ]

    def _init_layers(self):
        layers = self._display.environment.hydrography.loaded_layers
        for i, layer in enumerate(layers):
            rank = layer.z_order + i
            bins = len(self._display.environment.scope.depths)
            color = color_picker(i, bins)
            artist = self.new_artist(layer.geometry, color, rank)
            self._seabeds[rank] = artist
        shore = self._display.environment.topography.shore
        color = color_picker(shore.color)
        self._shore = self.new_artist(shore.geometry, color, shore.z_order)
        land = self._display.environment.topography.land
        color = color_picker(land.color)
        self._land = self.new_artist(land.geometry, color, land.z_order)
        if self._display.environment.scope.border:
            center = self._display.environment.scope.extent.center
            size = self._display.environment.scope.extent.size
            geometry = spl.Rectangle(*center, width=size[0] / 2, heading=0, height=size[1] / 2).geometry
            color = (color_picker("black")[0], "none")
            self.new_artist(geometry, color, 10000, linewidth=3)

    def new_artist(self, geometry, color, z_order=None, **kwargs):
        kwargs["crs"] = self._display.crs
        if z_order is not None:
            kwargs["zorder"] = z_order
        if isinstance(color, tuple):
            kwargs["ec"] = color[0]
            kwargs["fc"] = color[1]
        else:
            kwargs["color"] = color
        artist = self._display.axes.add_feature(ShapelyFeature([geometry], **kwargs))
        if z_order is None:
            artist.set_animated(True)
        return artist

    def add_arrow(self, start, end, color_name, buffer, fill, head_size, linewidth, linestyle):
        if buffer is None:
            buffer = 5
        if head_size is None:
            head_size = 50
        body = spl.Arrow(start=start, end=end, width=buffer).body(head_size)
        self.add_overlay(body, color_name, fill, linewidth, linestyle)

    def add_circle(self, center, radius, color_name, fill, linewidth, linestyle, alpha):
        geometry = spl.Circle(*center, radius).geometry
        self.add_overlay(geometry, color_name, fill, linewidth, linestyle, alpha)

    def add_line(self, points, color_name, buffer, linewidth, linestyle, marker):
        if buffer is None:
            buffer = 5
        if buffer == 0:
            x_coordinates, y_coordinates = zip(*points)
            self._display.axes.plot(
                x_coordinates, y_coordinates, color=color_picker(color_name)[0], linewidth=linewidth, linestyle=linestyle, marker=marker, transform=self._display.crs
            )
        else:
            geometry = spl.Line(points=points).geometry.buffer(buffer)
            self.add_overlay(geometry, color_name, True, linewidth, linestyle)

    def add_polygon(self, shape, color, interiors, fill, linewidth, linestyle, alpha=1.0):
        try:
            if isinstance(shape, sgeo.MultiPolygon) or isinstance(shape, sgeo.GeometryCollection):
                shape = list(shape.geoms)
            else:
                shape = list(shape)
        except TypeError:
            shape = [shape]
        if isinstance(shape[0], tuple) or isinstance(shape[0], list):
            shape = [shape]
        for geometry in shape:
            geometry = spl.Area.new_polygon(geometry, interiors)
            self.add_overlay(geometry, color, fill, linewidth, linestyle, alpha)

    def add_rectangle(self, center, size, color_name, rotation, fill, linewidth, linestyle, alpha):
        geometry = spl.Rectangle(*center, heading=rotation, width=size[0], height=size[1]).geometry
        self.add_overlay(geometry, color_name, fill, linewidth, linestyle, alpha)

    def add_overlay(self, geometry, color_name, fill, linewidth, linestyle, alpha=1.0):
        color = color_picker(color_name)
        if fill is False:
            color = color[0], "none"
        kwargs = {}
        if linewidth is not None:
            kwargs["linewidth"] = linewidth
        if linestyle is not None:
            kwargs["linestyle"] = linestyle
        kwargs["alpha"] = alpha
        self.new_artist(geometry, color, 0, **kwargs)

    def update_waypoints(self, number, pick, coords=None):
        path = self._paths[number - 1]
        index = path.locate_waypoint(*pick)
        if coords is not None:
            path.add_waypoint(*coords, index)
        else:
            if index is None:
                index = path.locate_edge(*pick)
                if index is None:
                    return
                else:
                    path.add_waypoint(*pick, index, edge=True)
            else:
                path.remove_waypoint(index)
        if path.artist:
            path.artist.remove()
        color = color_picker(path.color)
        path.artist = self.new_artist(path.multi_shape, color)
        points = [(wp.x, wp.y) for wp in path.waypoints]
        file_path = [utils.files.path.path1, utils.files.path.path2][number - 1]
        utils.files.write_rows_to_csv(points, file_path)

    def update_ownship(self):
        if self.show_ownship:
            ownship = self._display.environment.ownship
            bc = color_picker("cyan")
            hc = color_picker("full_horizon")
            if self._ownship:
                self._ownship.remove()
            if self._horizon:
                self._horizon.remove()
            self._ownship = self.new_artist(ownship.geometry, bc)
            self._horizon = self.new_artist(ownship.horizon, hc)
            if not self.show_hazards:
                self._horizon.set_visible(False)

    def update_hazards(self):
        if self.show_hazards:
            ownship = self._display.environment.ownship
            safe_area = self._display.environment.safe_area
            sectors = self._display.environment.ownship.horizon_sectors
            static_points = [() for _ in range(len(sectors))]
            dynamic_points = [() for _ in range(len(sectors))]
            for i, (color, geometry) in enumerate(sectors.items()):
                for features in [self._hazards, self._arrows]:
                    artist = features.pop(color, None)
                    if artist:
                        artist.remove()
                vessel_horizons = spl.Shape.collect([v["ship"].horizon for v in self._vessels.values()])
                static = geometry.difference(safe_area.geometry)
                dynamic = geometry.intersection(vessel_horizons)
                if not (static.is_empty and dynamic.is_empty):
                    self._hazards[color] = self.new_artist(static.union(dynamic), color_picker(color))
                    if self.show_arrows:
                        arrow, length1 = None, -1
                        if not static.is_empty:
                            arrow1, length1 = self.closest(ownship, static)
                            static_points[i] = arrow1.exterior.coords[0]
                            arrow = arrow1
                        if not dynamic.is_empty:
                            arrow2, length2 = self.closest(ownship, dynamic)
                            dynamic_points[i] = arrow2.exterior.coords[0]
                            if arrow is None or length2 < length1:
                                arrow = arrow2
                        self._arrows[color] = self.new_artist(arrow, color_picker("orange"))
            utils.files.write_rows_to_csv(static_points, utils.files.path.static)
            utils.files.write_rows_to_csv(dynamic_points, utils.files.path.dynamic)

    @staticmethod
    def closest(ownship, hazards):
        if not spl.Shape.is_multi(hazards):
            hazards = spl.Shape.as_multi(hazards)
        lines = []
        for polygon in list(hazards):
            near = ownship.closest_points(polygon.exterior)
            lines.append(spl.Shape.line_between(near, ownship.center))
        shortest = sorted(lines, key=lambda x: x.length)[0]
        interpolated = shortest.interpolate(min(ownship.dimensions[1] * 0.8, shortest.length * 0.3))
        head, base = shortest.coords[0], interpolated.coords[0]
        (x1, y1), (x2, y2) = head, base
        dx, dy = (x2 - x1) / 3, (y2 - y1) / 3
        left, right = (x2 - dy, y2 + dx), (x2 + dy, y2 - dx)
        return spl.Shape.arrow_head([(x1, y1), right, left]), shortest.length

    def update_vessels(self):
        if self.show_vessels:
            entries = list(utils.files.read_ship_poses())
            if entries is not None:
                new_vessels = {}
                for ship_details in entries:
                    ship_id = ship_details[0]
                    pose = ship_details[1:4]
                    other = ship_details[4]
                    if len(other) > 0 and isinstance(other[0], str):
                        color = color_picker(other[0])
                    else:
                        color = color_picker("red")
                    kwargs = dict(
                        scale=float(other[1]) if len(other) > 1 else 1.0,
                        lon_scale=float(other[2]) if len(other) > 2 else 2.0,
                        lat_scale=float(other[3]) if len(other) > 3 else 1.0,
                    )
                    ship = spl.Ship(*pose, **kwargs)
                    artist = self.new_artist(ship.geometry, color)
                    if self._vessels.get(ship_id, None):
                        self._vessels.pop(ship_id)["artist"].remove()
                    new_vessels[ship_id] = dict(ship=ship, artist=artist)
                self.replace_vessels(new_vessels)

    def replace_vessels(self, new_artists):
        for vessel in self._vessels.values():
            vessel["artist"].remove()
        self._vessels = new_artists

    def toggle_vessels_visibility(self, new_state: bool = None):
        if new_state is None:
            new_state = not self.show_vessels
        self.show_vessels = new_state
        for vessel in self._vessels.values():
            artist = vessel["artist"]
            artist.set_visible(not artist.get_visible())
        self.update_vessels()
        self.update_hazards()
        self._display.update_plot()

    def toggle_topography_visibility(self, new_state: bool = None):
        if new_state is None:
            new_state = not self._land.get_visible()
        self._land.set_visible(new_state)
        self._shore.set_visible(new_state)
        self._display.draw_plot()

    def toggle_ownship_visibility(self):
        self.show_ownship = not self.show_ownship
        if self.show_ownship:
            self.update_ownship()
        self._ownship.set_visible(self.show_ownship)
        self.toggle_hazards_visibility()
        self._display.update_plot()

    def toggle_hazards_visibility(self):
        self.show_hazards = not self.show_hazards
        self._horizon.set_visible(self.show_hazards)
        for artist in self._hazards.values():
            artist.set_visible(self.show_hazards)
        self.toggle_arrows_visibility(self.show_hazards)

    def toggle_arrows_visibility(self, new_state: bool = None):
        if new_state is None:
            new_state = not self.show_arrows
            self.show_arrows = new_state
        for artist in self._arrows.values():
            artist.set_visible(new_state)
        self.update_hazards()
        self._display.update_plot()

    def show_top_hidden_layer(self):
        artists = self._z_sorted_seabeds(descending=False)
        self._toggle_next_visibility_layer(artists, visibility=True)

    def hide_top_visible_layer(self):
        artists = self._z_sorted_seabeds(descending=False)
        self._toggle_next_visibility_layer(artists, visibility=False)

    def hide_bottom_visible_layer(self):
        artists = self._z_sorted_seabeds(descending=True)
        self._toggle_next_visibility_layer(artists, visibility=False)

    def show_bottom_hidden_layer(self):
        artists = self._z_sorted_seabeds(descending=True)
        self._toggle_next_visibility_layer(artists, visibility=True)

    def _toggle_next_visibility_layer(self, artists, visibility):
        if self._any_toggleable_layer(artists, not visibility):
            artist = self._next_visibility_layer(artists, not visibility)
            artist.set_visible(visibility)
            self._display.draw_plot()

    def _z_sorted_seabeds(self, descending=False):
        artist_keys = sorted(self._seabeds, reverse=descending)
        return [self._seabeds[key] for key in artist_keys]

    @staticmethod
    def _any_toggleable_layer(artists, visible):
        return any([a.get_visible() is visible for a in artists])

    @staticmethod
    def _next_visibility_layer(artists, visibility):
        return next(a for a in artists if a.get_visible() is visibility)

    @staticmethod
    def vessels_to_file(vessel_poses: list):
        utils.files.write_rows_to_csv([("id", "easting", "northing", "heading", "color")] + vessel_poses, utils.files.path.vessels)



from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from typing import List, Tuple

from shapely import affinity
from shapely import geometry as geo

from . import base


@dataclass
class Area(base.Shape):
    geometry: geo.Polygon = field(default_factory=geo.Polygon)

    @staticmethod
    def new_polygon(exterior, interiors=None):
        return geo.Polygon(exterior, interiors)


@dataclass
class Line(base.Shape):
    points: List[Tuple[float, float]] = None

    def __post_init__(self):
        if self.points is None or len(self.points) < 2:
            raise ValueError(
                f"{self.__class__.__name__} must contain at least "
                f"2 pairs of coordinates"
            )
        self.geometry = geo.LineString(self.points)


@dataclass
class Arrow(base.Shape):
    start: Tuple[float, float] = None
    end: Tuple[float, float] = None
    width: float = None

    def __post_init__(self):
        if self.start is None or self.end is None:
            raise ValueError(
                f"{self.__class__.__name__} must have a start and an end point"
            )
        self.geometry = geo.LineString((self.start, self.end))

    @property
    def vector(self) -> Tuple[float, float]:
        return self.end[0] - self.start[0], self.end[1] - self.start[1]

    def body(self, head_size):
        if not head_size >= 0:
            raise ValueError(
                f"{self.__class__.__name__} " "should have a non-negative head size"
            )
        length = self.geometry.length
        arrow_head_length = max(length - head_size, 0)
        x1, y1 = self.start
        x2, y2 = self.geometry.interpolate(arrow_head_length).coords[0]
        unit = self.vector[0] / length, self.vector[1] / length
        dx1, dy1 = -unit[1] * self.width, unit[0] * self.width
        dx2, dy2 = dx1 * 3, dy1 * 3
        tip_left, tip_right = (x2 + dx2, y2 + dy2), (x2 - dx2, y2 - dy2)
        base_left, base_right = (x2 + dx1, y2 + dy1), (x2 - dx1, y2 - dy1)
        start_left, start_right = (x1 + dx1, y1 + dy1), (x1 - dx1, y1 - dy1)
        return geo.Polygon(
            (
                self.end,
                tip_left,
                base_left,
                start_left,
                start_right,
                base_right,
                tip_right,
            )
        )


@dataclass
class Circle(Area, base.Radial, base.Coordinates):
    def __post_init__(self):
        if self.radius <= 0:
            raise ValueError(f"{self.__class__.__name__} should have a positive area")
        self.center = geo.Point(self.x, self.y)
        self.geometry = geo.Polygon(self.center.buffer(self.radius))


@dataclass
class Body(Area, base.Oriented, base.Coordinates):
    def __post_init__(self):
        self.center = geo.Point(self.x, self.y)
        self.geometry = self.rotate(self._body_polygon())

    def _body_polygon(self) -> geo.Polygon:
        raise NotImplementedError

    def rotate(self, polygon):
        return affinity.rotate(
            polygon,
            -self.heading,
            use_radians=not self.in_degrees,
            origin=(self.center.x, self.center.y),
        )


@dataclass
class Rectangle(Body):
    width: float = 0.0
    height: float = 0.0

    def _body_polygon(self) -> geo.Polygon:
        if not self.width > 0 or not self.height > 0:
            raise ValueError(f"{self.__class__.__name__} should have a positive area")
        return geo.Polygon(
            (
                (self.x - self.width, self.y - self.height),
                (self.x + self.width, self.y - self.height),
                (self.x + self.width, self.y + self.height),
                (self.x - self.width, self.y + self.height),
            )
        )


@dataclass
class Ship(Body):
    dimensions = 16, 80
    scale: float = 1.0
    lon_scale: float = 10.0
    lat_scale: float = 10.0

    def _body_polygon(self) -> geo.Polygon:
        x, y = self.x, self.y
        w, h = (d * self.scale for d in self.dimensions)
        x_min, x_max = x - w / 2, x + w / 2
        y_min, y_max = y - h / 2, y + h / 2 - w
        left_aft, right_aft = (x_min, y_min), (x_max, y_min)
        left_bow, right_bow = (x_min, y_max), (x_max, y_max)
        coords = [left_aft, left_bow, (x, y + h / 2), right_bow, right_aft]
        return geo.Polygon(coords)

    @property
    def parameters(self):
        return (
            self.x,
            self.y,
            self.heading,
            self.scale,
            self.lon_scale,
            self.lat_scale,
        )

    @property
    def head(self):
        return self.x, self.y + 200 * self.lon_scale

    @property
    def starboard_bow(self):
        return self.x + 100 * self.lat_scale, self.y + 80 * self.lon_scale

    @property
    def starboard_side(self):
        return self.x + 100 * self.lat_scale, self.y - 40 * self.lon_scale

    @property
    def starboard_aft(self):
        return self.x + 50 * self.lat_scale, self.y - 100 * self.lon_scale

    @property
    def port_aft(self):
        return self.x - 50 * self.lat_scale, self.y - 100 * self.lon_scale

    @property
    def port_side(self):
        return self.x - 100 * self.lat_scale, self.y - 40 * self.lon_scale

    @property
    def port_bow(self):
        return self.x - 100 * self.lat_scale, self.y + 80 * self.lon_scale

    @property
    def horizon(self) -> geo.Polygon:
        return self.rotate(
            geo.Polygon(
                [
                    self.head,
                    self.starboard_bow,
                    self.starboard_side,
                    self.starboard_aft,
                    self.port_aft,
                    self.port_side,
                    self.port_bow,
                ]
            )
        )

    @property
    def starboard_bow_horizon(self) -> geo.Polygon:
        return self.rotate(
            geo.Polygon(
                [
                    self.center.coords[0],
                    self.head,
                    self.starboard_bow,
                ]
            )
        )

    @property
    def starboard_side_horizon(self) -> geo.Polygon:
        return self.rotate(
            geo.Polygon(
                [
                    self.center.coords[0],
                    self.starboard_bow,
                    self.starboard_side,
                ]
            )
        )

    @property
    def starboard_aft_horizon(self) -> geo.Polygon:
        return self.rotate(
            geo.Polygon(
                [
                    self.center.coords[0],
                    self.starboard_side,
                    self.starboard_aft,
                ]
            )
        )

    @property
    def rear_aft_horizon(self) -> geo.Polygon:
        return self.rotate(
            geo.Polygon(
                [
                    self.center.coords[0],
                    self.starboard_aft,
                    self.port_aft,
                ]
            )
        )

    @property
    def port_aft_horizon(self) -> geo.Polygon:
        return self.rotate(
            geo.Polygon(
                [
                    self.center.coords[0],
                    self.port_aft,
                    self.port_side,
                ]
            )
        )

    @property
    def port_side_horizon(self) -> geo.Polygon:
        return self.rotate(
            geo.Polygon(
                [
                    self.center.coords[0],
                    self.port_side,
                    self.port_bow,
                ]
            )
        )

    @property
    def port_bow_horizon(self) -> geo.Polygon:
        return self.rotate(
            geo.Polygon(
                [
                    self.center.coords[0],
                    self.port_bow,
                    self.head,
                ]
            )
        )

    @property
    def horizon_sectors(self) -> dict:
        return dict(
            port_bow=self.port_bow_horizon,
            port_side=self.port_side_horizon,
            port_aft=self.port_aft_horizon,
            rear_aft=self.rear_aft_horizon,
            starboard_aft=self.starboard_aft_horizon,
            starboard_side=self.starboard_side_horizon,
            starboard_bow=self.starboard_bow_horizon,
        )


@dataclass
class Waypoint(Area, base.Radial, base.Coordinates):
    resolution: InitVar[int] = 0

    def __post_init__(self, resolution):
        self.center = geo.Point(self.x, self.y)
        self.geometry = self.center.buffer(self.radius, resolution=resolution)

    def contains(self, x, y):
        return self.geometry.contains(geo.Point(x, y))


class Path:
    def __init__(self, color):
        self.color = color
        self.waypoints = []
        self.edges = []
        self.artist = None

    @property
    def multi_shape(self):
        return base.Shape.collect([x.geometry for x in self.waypoints] + self.edges)

    def add_waypoint(self, x, y, index=None, edge=False):
        radius = 30
        if index is None:
            prev_wp = self.waypoints[-1] if self.waypoints else None
            waypoint = Waypoint(x, y, radius, resolution=2)
            if prev_wp:
                prev_wp.next = waypoint
                edge = self.edge_between(prev_wp, waypoint)
                self.edges.append(edge)
            self.waypoints.append(waypoint)
        else:
            if not edge:
                if index > 0:
                    prev_wp = self.waypoints[index - 1]
                else:
                    prev_wp = None
                if index < len(self.waypoints) - 1:
                    next_wp = self.waypoints[index + 1]
                else:
                    next_wp = None
                waypoint = Waypoint(x, y, radius, resolution=2)
                self.waypoints.pop(index)
                self.waypoints.insert(index, waypoint)
                if prev_wp:
                    edge1 = self.edge_between(prev_wp, waypoint)
                    self.edges.pop(index - 1)
                    self.edges.insert(index - 1, edge1)
                if next_wp:
                    edge2 = self.edge_between(waypoint, next_wp)
                    self.edges.pop(index)
                    self.edges.insert(index, edge2)
            else:
                prev_wp = self.waypoints[index]
                if index < len(self.waypoints) - 1:
                    next_wp = self.waypoints[index + 1]
                else:
                    next_wp = None
                waypoint = Waypoint(x, y, radius, resolution=2)
                self.waypoints.insert(index + 1, waypoint)
                new_edge = self.edge_between(prev_wp, waypoint)
                self.edges.insert(index, new_edge)
                if next_wp:
                    self.edges[index + 1] = self.edge_between(waypoint, next_wp)

    def remove_waypoint(self, index):
        self.waypoints.pop(index)
        if index == 0:
            if self.edges:
                self.edges.pop(0)
        elif index < len(self.waypoints):
            self.edges.pop(index)
            previous_wp = self.waypoints[index - 1]
            next_wp = self.waypoints[index]
            self.edges[index - 1] = self.edge_between(previous_wp, next_wp)
        else:
            self.edges.pop(-1)

    def locate_waypoint(self, x, y):
        for i, waypoint in enumerate(self.waypoints):
            if waypoint.contains(x, y):
                return i
        return None

    def locate_edge(self, x, y):
        for i, edge in enumerate(self.edges):
            if edge.contains(base.geo.Point(x, y)):
                return i
        return None

    @staticmethod
    def edge_between(wp1, wp2):
        line = wp1.line_between(wp1.center, wp2.center)
        return line.buffer(7, cap_style=2, join_style=3)



"""Contains functionality for reading, processing and validating seacharts configuration settings"""
from pathlib import Path
from typing import List

import yaml
from cerberus import Validator

from . import files
from . import paths as dcp  # default configuratin paths


class ENCConfig:
    """Class for maintaining Electronic Navigational Charts configuration settings"""

    def __init__(self, config_file_name: Path = dcp.config, **kwargs):

        self._schema = read_yaml_into_dict(dcp.config_schema)
        self.validator = Validator(self._schema)
        self._valid_sections = self.extract_valid_sections()

        self.parse(config_file_name)
        self.override(**kwargs)

    @property
    def settings(self):
        return self._settings

    @settings.setter
    def settings(self, new_settings: dict):
        self._settings = new_settings

    def extract_valid_sections(self) -> List[str]:
        if self._schema is None:
            raise ValueError("No configuration schema provided!")

        sections = []
        for section in self._schema.keys():
            sections.append(section)
        return sections

    def validate(self, settings: dict) -> None:
        if not settings:
            raise ValueError("Empty settings!")

        if not self._schema:
            raise ValueError("Empty schema!")

        if not self.validator.validate(settings):
            raise ValueError(f"Cerberus validation Error: {self.validator.errors}")

        self._settings["enc"]["depths"].sort()

        for file_name in self._settings["enc"]["files"]:
            files.verify_directory_exists(file_name)

    def parse(self, file_name=dcp.config) -> None:
        self._settings = read_yaml_into_dict(file_name)
        self.validate(self._settings)

    def override(self, section="enc", **kwargs) -> None:
        if not kwargs:
            return

        if section not in self._valid_sections:
            raise ValueError("Override settings in non-existing section!")

        new_settings = self._settings
        for key, value in kwargs.items():
            new_settings[section][key] = value

        self.validate(new_settings)

        self._settings = new_settings


def read_yaml_into_dict(file_name=dcp.config) -> dict:
    with open(file_name, encoding="utf-8") as config_file:
        output_dict = yaml.safe_load(config_file)
    return output_dict


def parse_key(key, defaults):
    """Returns default config parameter value for a given key, if it exists.

    Args:
        key (str): Key (parameter name) to parse
        defaults (dict): Dictionary containing default configuration values.

    Raises:
        ValueError: On missing or wrong input key

    Returns:
        Value: Value of parameter in config at the given key
    """
    default = defaults.get(key, None)
    if default is None:
        raise ValueError(
            f"Missing input parameter: '{key}': was not provided, "
            "and could not located in the current configuration file."
        )
    return default


def validate_key(key, value, v_type, sub_type=None, length=None) -> None:
    """Validates type of value at given key to be equal to v_type, or sub_type if value is a list/tuple of len = length.

    Args:
        key (str): Name of parameter to check value for
        value (_type_): Value of parameter in config at the given key
        v_type (_type_): Type of value
        sub_type (_type_, optional): Subtype if value is a list/tuple. Defaults to None.
        length (_type_, optional): Length if value is list/tuple. Defaults to None.

    Raises:
        ValueError: On wrong type of value at the given key.
    """
    if isinstance(value, list) or isinstance(value, tuple):
        if not all([isinstance(v, sub_type) for v in value]):
            raise ValueError("Invalid input format: " f"'{key}' should be a {v_type.__name__} of {sub_type}.")
        if length is not None and len(value) != length:
            raise ValueError("Invalid input format: " f"'{key}' should be a {v_type.__name__} of length {length}.")
    else:
        if not isinstance(value, v_type):
            raise ValueError("Invalid input format: " f"'{key}' should have type {v_type}.")


# def save(kwargs, file_name) -> None:
#     """Saves configuration with values given by kwargs, in file file_name.

#     Args:
#         file_name (str): Absolute path to the configuration file
#         kwargs (dict): Dictionary representing the configuration
#     """
#     config = configparser.ConfigParser()
#     user_strings = {}
#     for key, value in kwargs.items():
#         if isinstance(value, tuple) or isinstance(value, list):
#             string = ', '.join([str(v) for v in value])
#         else:
#             string = str(value)
#         user_strings[key] = string
#     config['USER'] = user_strings
#     with open(file_name, 'w', encoding='utf8') as configfile:
#         config.write(configfile)



import matplotlib.colors as clr
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.cm import ScalarMappable


def _blues(bins=9):
    # noinspection PyUnresolvedReferences
    return plt.get_cmap('Blues')(np.linspace(0.6, 1.0, bins))


def _greens(bins=9):
    # noinspection PyUnresolvedReferences
    return plt.get_cmap('Greens')(np.linspace(0.3, 0.9, bins))


_ship_colors = dict(
    red=('#ff0000', '#ff000055'),
    blue=('#0000ff', '#0000ff55'),
    green=('#00ff00', '#00ff0055'),
    yellow=('#ffff00', '#ffff0055'),
    cyan=('#00ffff', '#00ffff55'),
    magenta=('#ff00ff', '#ff00ff55'),
    pink=('#ff88ff', '#ff88ff55'),
    purple=('#bb22ff', '#bb22ff55'),
    orange=('#ff9900', '#ff990055'),
    white=('#ffffff', '#ffffff77'),
    lightgrey=('#b7b7b7', '#b7b7b755'),
    grey=('#666666', '#66666655'),
    darkgrey=('#333333', '#33333355'),
    black=('#000000', '#00000077'),
)

_horizon_colors = dict(
    full_horizon=('#ffffff55', '#ffffff11'),
    starboard_bow=('#00ff0099', '#00ff0055'),
    starboard_side=('#33ff3399', '#33ff3355'),
    starboard_aft=('#ccffcc99', '#ccffcc55'),
    rear_aft=('#eeeeee99', '#eeeeee55'),
    port_aft=('#ffcccc99', '#ffcccc55'),
    port_side=('#ff333388', '#ff333355'),
    port_bow=('#ff000066', '#ff000055'),
)

_layer_colors = dict(
    seabed=_blues()[0],
    land=_greens()[4],
    shore=_greens()[3],
    highlight=('#ffffff44', '#ffffff44'),
    blank=('#ffffffff', '#ffffffff'),
)


def color_picker(name, bins=None):
    if isinstance(name, int):
        return _blues(bins)[name]
    elif name in _ship_colors:
        return _ship_colors[name]
    elif name in _horizon_colors:
        return _horizon_colors[name]
    elif name in _layer_colors:
        return _layer_colors[name]
    elif name in clr.CSS4_COLORS:
        return clr.CSS4_COLORS[name]
    else:
        raise ValueError(
            f"{name} is not a valid color"
        )


def colorbar(axes, depths):
    depths = list(depths)
    ocean = list(_blues(len(depths)))
    colors = [_layer_colors['shore']] + ocean[:-1]
    c_map = clr.LinearSegmentedColormap.from_list(
        'Custom terrain', colors, len(colors)
    )
    c_map.set_over(ocean[-1])
    c_map.set_under(_layer_colors['land'])
    norm = clr.BoundaryNorm([0, 1] + depths[1:], c_map.N)
    kwargs = dict(extend='both', use_gridspec=True, extendfrac=0.1,
                  format='%1i m', ticks=[0, 1] + depths[1:],
                  boundaries=([depths[0] - 100] + [0, 1] +
                              depths[1:] + [depths[-1] + 100]))
    mappable = ScalarMappable(norm=norm, cmap=c_map)
    cb = plt.colorbar(mappable, cax=axes, **kwargs)
    cb.ax.tick_params(labelsize=20, length=5, width=1.5)
    cb.outline.set_linewidth(1.5)
    cb.ax.invert_yaxis()
    return cb



"""Contains file/directory related utility functions, such as functions for writing to csv files."""
import csv

from . import paths as path


def verify_directory_exists(dir_path) -> None:
    if not (path.external / dir_path).is_dir():
        raise FileNotFoundError(
            f"Folder {dir_path} not found at:\r\n{path.external}."
        )


def build_directory_structure(features=None) -> None:
    if features is None:
        path.data.mkdir(exist_ok=True)
        path.reports.mkdir(exist_ok=True)

        path.external.mkdir(exist_ok=True)
        path.hazards.mkdir(exist_ok=True)
        path.paths.mkdir(exist_ok=True)
        path.shapefiles.mkdir(exist_ok=True)

        path.vessels.touch(exist_ok=True)
        path.dynamic.touch(exist_ok=True)
        path.static.touch(exist_ok=True)
        path.path1.touch(exist_ok=True)
        path.path2.touch(exist_ok=True)
    else:
        for feature in features:
            shapefile_dir = path.shapefiles / feature.lower()
            shapefile_dir.mkdir(parents=True, exist_ok=True)


def write_rows_to_csv(rows, file_path) -> None:
    with open(file_path, 'w') as csv_file:
        writer = csv.writer(csv_file, delimiter=',', lineterminator='\n')
        writer.writerows(rows)


def read_ship_poses():
    try:
        with open(path.vessels) as csv_file:
            reader = csv.reader(csv_file, delimiter=',')
            _ = next(reader, None)
            rows = tuple(reader)
    except (PermissionError or FileNotFoundError
            or StopIteration or RuntimeError):
        return
    for row in rows:
        if row:
            yield int(row[0]), int(row[1]), int(row[2]), float(row[3]), row[4:]



from pathlib import Path
from typing import Any, List, Tuple, Union

import matplotlib
from cartopy.crs import UTM

import seacharts.display as dis
import seacharts.environment as env
import seacharts.utils as utils


class ENC:
    """Electronic Navigational Charts

    Reads and extracts features from a user-specified region of spatial data
    given in Cartesian coordinates. Based on Matplotlib, Shapely and Cartopy.
    An independent visualization window may be spawned and displayed using the
    multiprocessing option. Geometric shapes may be accessed through the
    attributes 'land', 'shore', and 'seabed'.

    :param config_file: string containing path to configuration file
    :param multiprocessing: bool for independent visualization display
    :param kwargs: Includes the following optional parameters:
        :param size: tuple(width, height) of bounding box size
        :param origin: tuple(easting, northing) box origin of coordinates
        :param center: tuple(easting, northing) box center of coordinates
        :param buffer: int of dilation or erosion distance for geometries
        :param tolerance: int of maximum tolerance distance for geometries
        :param layers: list(str...) of feature layers to load or show
        :param depths: list(int...) of depth bins for feature layers
        :param files: list(str...) of file names for zipped FGDB files
        :param new_data: bool indicating if new files should be parsed
        :param border: bool for showing a border around the environment plot
        :param verbose: bool for status printing during geometry processing
    """

    def __init__(
        self,
        config_file: Path = utils.paths.config,
        multiprocessing=False,
        **kwargs
    ):
        self._setup_matplotlib_parameters()
        if multiprocessing:
            dis.Display.init_multiprocessing()
            return

        self._cfg = utils.config.ENCConfig(config_file, **kwargs)

        self._environment = env.Environment(self._cfg.settings)
        self.land = self._environment.topography.land
        self.shore = self._environment.topography.shore
        self.seabed = self._environment.hydrography.bathymetry
        self._display = dis.Display(self._cfg.settings, self._environment)

    @property
    def size(self) -> Tuple[int, int]:
        """
        :return: tuple of bounding box size
        """
        return self._environment.scope.extent.size

    @property
    def center(self) -> Tuple[int, int]:
        """
        :return: tuple of ENC center coordinates
        """
        return self._environment.scope.extent.center

    @property
    def origin(self) -> Tuple[int, int]:
        """
        :return: tuple of ENC origin (lower left) coordinates.
        """
        return self._environment.scope.extent.origin

    @property
    def bbox(self) -> Tuple[int, int, int, int]:
        """
        :return: tuple of bounding box coordinates (xmin, ymin, xmax, ymax)
        """
        return self._environment.scope.extent.bbox

    @property
    def considered_depths(self) -> List[int]:
        """
        :return: list of considered depth bins
        """
        return self._environment.scope.depths

    @property
    def utm_zone(self) -> int:
        """
        :return: integer of UTM zone number
        """
        return self._environment.scope.extent.utm_zone

    @property
    def figure_active(self) -> bool:
        """
        :return: boolean indicating if a figure is active
        """
        return self._display.is_active

    @property
    def crs(self) -> UTM:
        """Return the coordinate reference system projection used, as UTM object."""
        if self._display.crs is None:
            return UTM(self.utm_zone)
        return self._display.crs

    @property
    def supported_crs(self) -> str:
        """Return the supported coordinate reference system, as string."""
        return self._environment.supported_crs

    @property
    def supported_layers(self) -> str:
        """Return the supported feature layers."""
        return self._environment.supported_layers

    def fullscreen_mode(self, arg: bool = True) -> None:
        """
        Enable or disable fullscreen mode view of environment figure.
        :param arg: boolean switching fullscreen mode on or off
        :return: None
        """
        self._display.toggle_fullscreen(arg)

    def colorbar(self, arg: bool = True) -> None:
        """
        Enable or disable the colorbar legend of environment figure.
        :param arg: boolean switching the colorbar on or off.
        """
        self._display.toggle_colorbar(arg)

    def dark_mode(self, arg: bool = True) -> None:
        """
        Enable or disable dark mode view of environment figure.
        :param arg: boolean switching dark mode on or off
        :return: None
        """
        self._display.toggle_dark_mode(arg)

    def add_vessels(self, *args: Tuple[int, int, int, int, str]) -> None:
        """
        Add colored vessel features to the displayed environment plot.
        :param args: tuples with id, easting, northing, heading, color
        :return: None
        """
        self._display.refresh_vessels(list(args))

    def clear_vessels(self) -> None:
        """
        Remove all vessel features from the environment plot.
        :return: None
        """
        self._display.refresh_vessels([])

    def add_ownship(
        self,
        easting: int,
        northing: int,
        heading: float,
        hull_scale: float = 1.0,
        lon_scale: float = 10.0,
        lat_scale: float = 10.0,
    ) -> None:
        """
        Add the body of a controllable vessel to the environment.
        :param easting: int denoting the ownship X coordinate
        :param northing: int denoting the ownship Y coordinate
        :param heading: float denoting the ownship heading in degrees
        :param hull_scale: optional float scaling the ownship body size
        :param lon_scale: optional float scaling the longitudinal horizon
        :param lat_scale: optional float scaling the lateral horizon
        :return: None
        """
        self._environment.create_ownship(
            easting, northing, heading, hull_scale, lon_scale, lat_scale
        )
        self._display.update_plot()

    def remove_ownship(self) -> None:
        """
        Remove the controllable vessel from the environment.
        :return: None
        """
        self._environment.ownship = None

    def add_hazards(self, depth: int, buffer: int = 0) -> None:
        """
        Add hazardous areas to the environment, filtered by given depth.
        :param depth: int denoting the filter depth
        :param buffer: optional int denoting the buffer distance
        :return: None
        """
        self._environment.filter_hazardous_areas(depth, buffer)

    def draw_arrow(
        self,
        start: Tuple[float, float],
        end: Tuple[float, float],
        color: str,
        width: float = None,
        fill: bool = False,
        head_size: float = None,
        thickness: float = None,
        edge_style: Union[str, tuple] = None,
    ) -> None:
        """
        Add a straight arrow overlay to the environment plot.
        :param start: tuple of start point coordinate pair
        :param end: tuple of end point coordinate pair
        :param color: str of line color
        :param width: float denoting the line buffer width
        :param fill: bool which toggles the interior arrow color on/off
        :param thickness: float denoting the Matplotlib linewidth
        :param edge_style: str or tuple denoting the Matplotlib linestyle
        :param head_size: float of head size (length) in meters
        :return: None
        """
        self._display.features.add_arrow(
            start, end, color, width, fill, head_size, thickness, edge_style
        )

    def draw_circle(
        self,
        center: Tuple[float, float],
        radius: float,
        color: str,
        fill: bool = True,
        thickness: float = None,
        edge_style: Union[str, tuple] = None,
        alpha: float = 1.0,
    ) -> None:
        """
        Add a circle or disk overlay to the environment plot.
        :param center: tuple of circle center coordinates
        :param radius: float of circle radius
        :param color: str of circle color
        :param fill: bool which toggles the interior disk color
        :param thickness: float denoting the Matplotlib linewidth
        :param edge_style: str or tuple denoting the Matplotlib linestyle
        :param alpha: float denoting the Matplotlib alpha value
        :return: None
        """
        self._display.features.add_circle(
            center, radius, color, fill, thickness, edge_style, alpha
        )

    def draw_line(
        self,
        points: List[Tuple[float, float]],
        color: str,
        width: float = None,
        thickness: float = None,
        edge_style: Union[str, tuple] = None,
        marker_type: str = None,
    ) -> None:
        """
        Add a straight line overlay to the environment plot.
        :param points: list of tuples of coordinate pairs
        :param color: str of line color
        :param width: float denoting the line buffer width
        :param thickness: float denoting the Matplotlib linewidth
        :param edge_style: str or tuple denoting the Matplotlib linestyle
        :param marker_type: str denoting the Matplotlib marker type
        :return: None
        """
        self._display.features.add_line(
            points, color, width, thickness, edge_style, marker_type
        )

    def draw_polygon(
        self,
        geometry: Union[Any, List[Tuple[float, float]]],
        color: str,
        interiors: List[List[Tuple[float, float]]] = None,
        fill: bool = True,
        thickness: float = None,
        edge_style: Union[str, tuple] = None,
        alpha: float = 1.0,
    ) -> None:
        """
        Add an arbitrary polygon shape overlay to the environment plot.
        :param geometry: Shapely geometry or list of exterior coordinates
        :param interiors: list of lists of interior polygon coordinates
        :param color: str of rectangle color
        :param fill: bool which toggles the interior shape color
        :param thickness: float denoting the Matplotlib linewidth
        :param edge_style: str or tuple denoting the Matplotlib linestyle
        :param alpha: float denoting the Matplotlib alpha value
        :return: None
        """
        self._display.features.add_polygon(
            geometry, color, interiors, fill, thickness, edge_style, alpha
        )

    def draw_rectangle(
        self,
        center: Tuple[float, float],
        size: Tuple[float, float],
        color: str,
        rotation: float = 0.0,
        fill: bool = True,
        thickness: float = None,
        edge_style: Union[str, tuple] = None,
        alpha: float = 1.0,
    ) -> None:
        """
        Add a rectangle or box overlay to the environment plot.
        :param center: tuple of rectangle center coordinates
        :param size: tuple of rectangle (width, height)
        :param color: str of rectangle color
        :param rotation: float denoting the rectangle rotation in degrees
        :param fill: bool which toggles the interior rectangle color
        :param thickness: float denoting the Matplotlib linewidth
        :param edge_style: str or tuple denoting the Matplotlib linestyle
        :param alpha: float denoting the Matplotlib alpha value
        :return: None
        """
        self._display.features.add_rectangle(
            center, size, color, rotation, fill, thickness, edge_style, alpha
        )

    def start_display(self) -> None:
        """
        Start the ENC figure display.
        :return: None
        """
        self._display.start(self._cfg.settings)

    def show_display(self, duration: float = 0.0) -> None:
        """
        Show a Matplotlib display window of a maritime environment.
        :param duration: optional int for window pause duration
        :return: None
        """
        self._display.show(duration)

    def get_display_handle(self):
        """Returns figure and axes handles to the seacharts display."""
        return self._display.figure, self._display.axes

    def refresh_display(self) -> None:
        """
        Manually redraw the environment display window.
        :return: None
        """
        self._display.draw_plot()

    def close_display(self) -> None:
        """
        Close the environment display window and clear all vessels.
        :return: None
        """
        self._display.terminate()
        self.clear_vessels()

    def save_image(
        self,
        name: str = None,
        path: Path | None = None,
        scale: float = 1.0,
        extension: str = "png",
    ) -> None:
        """
        Save the environment plot as a .png image.
        :param name: optional str of file name
        :param path: optional Path of file path
        :param scale: optional float scaling the image resolution
        :param extension: optional str of file extension name
        :return: None
        """
        self._display.save_figure(name, path, scale, extension)

    @staticmethod
    def _setup_matplotlib_parameters() -> None:
        """
        Set matplotlib parameters used for the entire ENC package.
        :return: None
        """
        matplotlib.rcParams["pdf.fonttype"] = 42
        matplotlib.rcParams["ps.fonttype"] = 42
        matplotlib.use("TkAgg")



import warnings

import fiona

from . import paths as path


class ShapefileParser:
    def __init__(self, bounding_box, file_names, verbose):
        self.bounding_box = bounding_box
        self.file_names = file_names
        self.verbose = verbose

    def read_fgdb(self, label, external_labels, depth):
        for file_name in self.file_names:
            file_path = path.external / file_name
            records = self._parse_layers(file_path, external_labels, depth)
            yield from self._parse_records(records, label)

    def read_shapefile(self, label):
        file_path = self._shapefile_path(label)
        if file_path.exists():
            yield from self._read_spatial_file(file_path)

    def _parse_layers(self, file_path, external_labels, depth):
        for label in external_labels:
            if isinstance(label, dict):
                layer, depth_label = label['layer'], label['depth']
                records = self._read_spatial_file(file_path, layer=layer)
                for record in records:
                    if record['properties'][depth_label] >= depth:
                        yield record
            else:
                yield from self._read_spatial_file(file_path, layer=label)

    def _read_spatial_file(self, file_path, **kwargs):
        with fiona.open(file_path, 'r', **kwargs) as source:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=RuntimeWarning)
                for record in source.filter(bbox=self.bounding_box):
                    yield record
        return

    def _parse_records(self, records, label):
        for i, record in enumerate(records):
            if self.verbose:
                print(f"\rNumber of {label} records read: {i + 1}", end='')
            yield record
        return

    def write(self, shape):
        geometry = shape.mapping
        file_path = self._shapefile_path(shape.label)
        with self.writer(file_path, geometry['type']) as sink:
            sink.write(self._as_record(shape.depth, geometry))

    def writer(self, file_path, geometry_type):
        return fiona.open(
            file_path, 'w',
            schema=self._as_record('int', geometry_type),
            driver='ESRI Shapefile', crs={'init': 'epsg:25833'}
        )

    @staticmethod
    def _depth_filter(depth_label, minimum_depth):
        return lambda r: r['properties'][depth_label] >= minimum_depth

    @staticmethod
    def _as_record(depth, geometry):
        return {'properties': {'depth': depth}, 'geometry': geometry}

    @staticmethod
    def _shapefile_path(label):
        return path.shapefiles / label / (label + '.shp')



from __future__ import annotations

from dataclasses import dataclass

from . import base


@dataclass
class Seabed(base.SingleDepthRegions):
    color = 'seabed'
    z_order = -300
    _external_labels = [
        dict(layer='dybdeareal', depth='minimumsdybde'),
        dict(layer='grunne', depth='dybde'),
    ]


@dataclass
class Land(base.ZeroDepthRegions):
    color = 'land'
    z_order = -100
    _external_labels = ['landareal']


@dataclass
class Shore(base.ZeroDepthRegions):
    color = 'shore'
    z_order = -200
    _external_labels = [
        'skjer', 'torrfall', 'landareal', 'ikkekartlagtsjomaltomr'
    ]


supported_layers = [cls.__name__ for cls in [
    Seabed,
    Land,
    Shore,
]]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List

import seacharts.spatial as spl
import seacharts.utils as utils

from .extent import Extent


@dataclass
class Scope:
    def __init__(self, settings: dict, extent: Extent):
        self.extent: Extent = extent
        self.buffer: int = settings["enc"]["buffer"]
        self.tolerance: int = settings["enc"]["tolerance"]
        self.layers: list = settings["enc"]["layers"]
        self.depths: list = settings["enc"]["depths"]
        self.files: list = settings["enc"]["files"]
        self.new_data: bool = settings["enc"]["new_data"]
        self.raw_data: bool = settings["enc"]["raw_data"]
        self.border: bool = settings["enc"]["border"]
        self.verbose: bool = settings["enc"]["verbose"]

        utils.files.build_directory_structure()

        seabed = spl.supported_layers[0].lower()
        if seabed in self.layers:
            self.layers.remove(seabed)
            for depth in self.depths:
                self.layers.append(f"{seabed}{depth}m")

        utils.files.build_directory_structure(self.layers)

        self.parser: utils.parser.ShapefileParser = utils.ShapefileParser(self.extent.bbox, self.files, self.verbose)



import pathlib

# Defines (hard-coded paths to relevant files)
root = pathlib.Path(__file__).parents[2]
package = root / "seacharts"

config = package / "config.yaml"
config_schema = package / "config_schema.yaml"

data = pathlib.Path.cwd() / "data"
external = data / "external"
shapefiles = data / "shapefiles"

vessels = data / "vessels.csv"

hazards = data / "hazards"
dynamic = hazards / "dynamic.csv"
static = hazards / "static.csv"

paths = data / "paths"
path1 = paths / "path1.csv"
path2 = paths / "path2.csv"

reports = root / "reports"
frames_dir = reports / "frames"
simulation = reports / "simulation.gif"
frame_files = frames_dir / "frame_*.png"
